from omigo_core import utils, dataframe, tsv, tsvutils
from omigo_hydra import file_paths_data_reader, file_paths_util, s3io_wrapper

def save_to_file(xtsv, output_file_name, s3_region = None, aws_profile = None):
    # do some validation
    xtsv = xtsv.validate()

    # check if it is a local file or s3
    if (output_file_name.startswith("s3://") == False):
        file_paths_util.create_local_parent_dir(output_file_name)

    # construct output file
    output_file = file_io_wrapper.TSVFileWriter(s3_region, aws_profile)
    
    # write
    output_file.save(xtsv, output_file_name)
    
    # debug
    utils.debug("save_to_file: file saved to: {}, num_rows: {}, num_cols: {}".format(output_file_name, xtsv.num_rows(), xtsv.num_cols()))

def check_exists(xtsv, s3_region = None, aws_profile = None):
    return file_paths_util.check_exists(xtsv, s3_region, aws_profile)

def read(input_file_or_files, sep = None, def_val_map = None, username = None, password = None, num_par = 0, s3_region = None, aws_profile = None):
    # convert the input to array
    input_files = utils.get_argument_as_array(input_file_or_files)

    # tasks 
    tasks = []

    def __read_inner__(input_file):
        # read file content                               
        lines = file_paths_util.read_file_content_as_lines(input_file, s3_region, aws_profile)
                                                          
        # take header and dat
        header = lines[0]
        data = lines[1:]
            
        # check if a custom separator is defined
        if (sep is not None):
            # check for validation
            for line in lines:
                if ("\t" in line):
                    raise Exception("Cant parse non tab separated file as it contains tab character:", input_file)

            # create header and data
            header = header.replace(sep, "\t")
            data = [x.replace(sep, "\t") for x in data]

        # return
        return dataframe.from_tsv(tsv.TSV(header, data))

    # create tasks
    for input_file in input_files:
        # check if it is a file or url
        if (input_file.startswith("http")):
            tsv_list.append(read_url_as_tsv(input_file, username = username, password = password))
            tasks.append(utils.ThreadPoolTask(read_url_as_tsv, input_file, username = username, password = password))
        else:
            tasks.append(utils.ThreadPoolTask(__read_inner__, input_file))

    # get result
    tsv_list = utils.run_with_thread_pool(tasks, num_par = num_par, wait_sec = 1)

    # merge and return
    return tsvutils.merge(tsv_list, def_val_map = def_val_map)

def __read_with_filter_transform_select_func__(cols):
    # create a inner function
    def __read_with_filter_transform_select_func_inner__(mp):
        result_mp = {}
        for c in cols:
            if (c in mp.keys()):
                result_mp[c] = str(mp[c])

        # return
        return result_mp

    return __read_with_filter_transform_select_func_inner__

def read_with_filter_transform(input_file_or_files, sep = None, def_val_map = None, filter_transform_func = None, cols = None, transform_func = None, s3_region = None, aws_profile = None):
    # check if cols is defined
    if (filter_transform_func is not None and cols is not None):
        raise Exception("tsvutils: read_with_filter_transform: either of filter_transform_func or cols parameter can be used")

    # use the map function for cols
    if (cols is not None and len(cols) > 0):
        filter_transform_func = __read_with_filter_transform_select_func__(cols)

    # check if filter_transform_func is defined
    if (filter_transform_func is None):
        xtsv = read(input_file_or_files, sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)

        # apply transform_func if defined
        xtsv_transform = transform_func(xtsv) if (transform_func is not None) else xtsv

        # return
        return xtsv_transform
    else:
        # resolve input
        input_files = utils.get_argument_as_array(input_file_or_files)

        # initialize result
        tsv_list = []

        # common keys
        common_keys = {}

        # iterate over all input files
        for input_file in input_files:
            # read the file
            x = read(input_file, sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)

            # update the common
            for h in x.get_header_fields():
                if (h not in common_keys.keys()):
                    common_keys[h] = 0
                common_keys[h] = common_keys[h] + 1

            # gather maps of maps
            result_maps = []
            keys = {}

            # iterate over the records of each map and generate a new one
            for mp in x.to_maps():
                mp2 = filter_transform_func(mp)
                if (mp2 is not None):
                    if (len(mp2) > 0):
                        result_maps.append(mp2)
                    for k in mp2.keys():
                        keys[k] = 1

            # check for empty maps
            if (len(keys) > 0):
                # output keys
                keys_sorted = []
                first_file = read(input_files[0], sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)
                for h in first_file.get_header_fields():
                    if (h in keys.keys()):
                        keys_sorted.append(h)

                # new header and data
                header2 = "\t".join(keys_sorted)
                data2 = []

                # iterate and generate header and data
                for mp in result_maps:
                    fields = []
                    for k in keys_sorted:
                        fields.append(mp[k])
                    data2.append("\t".join(fields))

                # debugging
                utils.trace("tsvutils: read_with_filter_transform: file read: {}, after filter num_rows: {}".format(input_file, len(data2)))

                # result tsv
                xtsv = tsv.TSV(header2, data2)

                # apply transformation function if defined
                xtsv_transform = transform_func(xtsv) if (transform_func is not None) else xtsv
                tsv_list.append(xtsv_transform)

        # Do a final check to see if all tsvs are empty
        if (len(tsv_list) > 0):
            # call merge on tsv_list
            return tsvutils.merge(tsv_list)
        else:
            # create an empty data tsv file with common header fields
            header_fields = []
            first_file = read(input_files[0], sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)
            for h in first_file.get_header_fields():
                if (common_keys[h] == len(input_files)):
                    header_fields.append(h)

            new_header = "\t".join(header_fields)
            new_data = []

            # return
            return tsv.TSV(new_header, new_data)

# TODO: replace this by etl_ext
def read_by_date_range(path, start_date_str, end_date_str, prefix, s3_region = None, aws_profile = None, granularity = "daily"):
    utils.warn_once("read_by_date_range: probably Deprecated")
    # read filepaths
    filepaths = file_paths_util.read_filepaths(path, start_date_str, end_date_str, prefix, s3_region, aws_profile, granularity)

    # check for headers validity
    if (file_paths_util.has_same_headers(filepaths, s3_region, aws_profile) == False):
        utils.warn("Mismatch in headers for different days. Choose the right date range: start: {}, end: {}".format(start_date_str, end_date_str))
        return None

    # read individual tsvs
    tsv_list = []
    for filepath in filepaths:
        x = read(filepath)
        tsv_list.append(x)

    # combine all together
    if (len(tsv_list) == 0):
        return None
    elif (len(tsv_list) == 1):
        return tsv_list[0]
    else:
        return tsv_list[0].union(tsv_list[1:])

def load_from_dir(path, start_date_str, end_date_str, prefix, s3_region = None, aws_profile = None, granularity = "daily"):
    # read filepaths
    filepaths = file_paths_util.read_filepaths(path, start_date_str, end_date_str, prefix, s3_region, aws_profile, granularity)
    return load_from_files(filepaths, s3_region, aws_profile)

def load_from_files(filepaths, s3_region, aws_profile):
    # check for headers validity
    if (file_paths_util.has_same_headers(filepaths, s3_region, aws_profile) == False):
        print("Invalid headers.")
        return None

    # initialize the file reader
    file_reader = file_paths_data_reader.FilePathsDataReader(filepaths, s3_region, aws_profile)

    # get header
    header = file_reader.get_header()
    data = []

    # get data
    while file_reader.has_next():
        # read next record
        line = file_reader.next()
        data.append(line)

    # close
    file_reader.close()

    return tsv.TSV(header, data)

def read_json_files_from_directories_as_tsv(paths, s3_region = None, aws_profile = None):
    # initialize fs
    fs = s3io_wrapper.S3FSWrapper(s3_region = s3_region, aws_profile = aws_profile)

    # result
    result = []

    # iterate through each directory
    for path in paths:
        # list all files
        files = fs.list_leaf_dir(path)

        # read file as set of lines
        for f in files:
            full_path = "{}/{}".format(path, f)
            lines = fs.read_file_contents_as_text(full_path).split("\n")

            # append to result
            result = result + lines

    # remove empty lines
    result = list(filter(lambda t: t.strip() != "", result))

    # create dataframe
    header_fields = ["json"]
    data_fields = list([t.split("\t") for t in result])
    df = dataframe.new_with_cols(header_fields, data = data_fields)

    # return
    return df




from omigo_core import tsv, utils

# ... 
# .select(["k1", "k2", "k3"]) \
# .topk("args1", "args2", kwarg1 = 100) \
# .randn(100) \
# .select_group_by_topk(["k1"], ["k2"], 13)
#
# .select(["k1", "k2", "k3"]) \
# .extend_class(CustomBlockTSV).custom_block("block_id1")

class CustomBlockTSV(tsv.TSV):
    OMIGO_CONFIG_RESOLUTION_PROPS = "OMIGO_CONFIG_RESOLUTION_PROPS"

    def __init__(self, header, data, base_config = None, config = None):
        super().__init__(header, data)
        self.base_config = base_config

    def apply_block(self, expected_cols = None, prefix = None, block_id = None, dmsg = ""):
        dmsg = utils.extend_inherit_message(dmsg, "apply_block")

        # base condition
        base_config_resolved = os.environ.get(OMIGO_CONFIG_RESOLUTION_PROPS, None) if (self.base_config is None) else self.base_config
        if (base_config_resolved is None):
            raise Exception("{}: base_config_resolved is None".format(dmsg))

        # check for config 
        if (block_id is None):
            utils.warn("{}: base_config_resolved: {}, block_id is None".format(dmsg, base_config_resolved))
            if (expected_cols is not None and len(expected_cols) > 0):
                return self.add_empty_cols_if_missing(expected_cols, prefix = prefix, dmsg = dmsg)
            else:
                return self 

        # read the config resolution properties
        json_obj = json.loads(tsvutils.read_file_contents_as_text(base_config))

        # read the config resolution class. TODO: reflection
        config_resolver = self.load_custom_config_resolver(json_obj)

        # read the block 
        block_config = config_resolver.get_block_config(block_id)

        # read operations
        operations = block_config["operations"]
        func_ops = block_config.__deserialize_operations__(operations)

        # iterate and apply
        for func_op in func_ops:
            print("noop. FIXME")

    def __load_custom_config_resolver__(self, json_obj):
        return CustomConfigResolution(json_obj)

class CustomConfigResolver:
    def __init__(self, json_obj):
        self.json_obj = json_obj

    def get_version(self):
        return json_obj["version"]

    def get_block_config(self, block_id, dmsg = ""):
        dmsg = utils.extends_inherit_message(dmsg, "CustomConfigResolver: get_block_config")

        # get all blocks
        blocks = self.json_obj["blocks"]

        # find and return
        if (block_id in blocks.keys()):
            return blocks[block_id]
        else:
            raise Exception("get_block_config: block_id: {} not found: {}".format(block_id, sorted(blocks.keys())))

    # op_list is a list of tuples - (name, args, kwargs)
    def __serialize_operations__(self, op_func_list, dmsg = ""):
        dmsg = utils.extends_inherit_message(dmsg, "CustomConfigResolver: __serialize_operations__")

        # use hydra to serialize
        ops = list([cluster_common_v2.ClusterInmemoryOperation(op).to_json() for op in op_func_list])
        return ops
            

    def __deserialize_operations__(self, op_json_list, dmsg = ""):
        dmsg = utils.extends_inherit_message(dmsg, "CustomConfigResolver: __deserialize_operations__")

        # use hydra to deserialize
        ops = list([cluster_common_v2.ClusterOperation.from_json(op) for op in op_json_list])
        return ops



import inspect
import importlib
import math
from omigo_core import tsv, utils

# TODO: Currying
# https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
def get_class_that_defined_method(method):
    # check if it is a method
    if inspect.ismethod(method):
        for cls in inspect.getmro(method.__self__.__class__):
            if method.__name__ in cls.__dict__:
                return cls

    # check if it is a function
    if inspect.isfunction(method):
        return getattr(inspect.getmodule(method), method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)

    raise Exception("get_class_that_defined_method: not able to find the class: {}".format(method)) 

# TODO: these methods are hacky version of class loading and referencing
def get_fully_qualified_class_name(clsref):
    # validation
    if (clsref is None):
        raise Exception("get_fully_qualified_class_name : clsref is None")

    # get function qualified name
    class_name = func.__name__

    # get the module name
    module_name = func.__module__

    # generate fully qualified name
    return "{}.{}".format(module_name, class_name)

# TODO: these methods are hacky version of class loading and referencing
def get_fully_qualified_name(func):
    # validation
    if (func is None):
        raise Exception("get_fully_qualified_name: func is None")

    # get function qualified name
    func_name = func.__qualname__

    # get the module name
    module_name = func.__module__

    # generate fully qualified name
    return "{}.{}".format(module_name, func_name)

# TODO: these methods are hacky version of class loading and referencing
def load_fully_qualified_func(fully_qual_func_name):
    # only omigo extensions are supported. TODO
    if (fully_qual_func_name.startswith("omigo_") == False):
        utils.debug("Non omigo extensions will not be supported in future: {}".format(fully_qual_func_name))

    # split into parts
    parts = fully_qual_func_name.split(".")

    # need fully qualified name
    if (len(parts) < 2):
        raise Exception("Not able to parse the function full qualified name: {}".format(fully_qual_func_name))
    
    # check for package or class based function hierarchy
    if (len(parts) == 2):
        # this is assumed to be package functions. TODO 
        utils.debug("Global functions may not be supported in future. Use their cluster_funcs counterparts")

        # assume its package.func_name
        module_name = parts[0]
        func_name = parts[1]

        # load method
        utils.debug("load_fully_qualified_func: {}, func_name: {}, module_name: {}".format(fully_qual_func_name, func_name, module_name))

        # check loading of the class 
        try:
            # load module
            m = importlib.import_module(module_name)

            # check for function
            return getattr(m, func_name)
        except Exception as e:
            utils.error("function not found: {}".format(fully_qual_func_name))
            raise e 
    else:
        # split function name from module name
        func_name = parts[-1]
        class_name = parts[-2]
        module_name = ".".join(parts[0:-2])

        # load method
        utils.debug("load_fully_qualified_func: {}, func_name: {}, class_name: {}, module_name: {}".format(fully_qual_func_name, func_name, class_name, module_name))

        # check loading of the class 
        try:
            # load module
            m = importlib.import_module(module_name)
            if (class_name in m.__dict__.keys()):
                # load class
                class_reference = m.__dict__[class_name]

                # check for function
                return getattr(class_reference, func_name)
            else:
                raise Exception("class name not found in the module: {}, class: {}".format(module_name, class_name)) 
        except Exception as e:
            utils.error("function not found: {}".format(fully_qual_func_name))
            raise e 

# TODO: these methods are hacky version of class loading and referencing
def verify_tsv_subclass_func(fully_qual_func_name):
    try:
        f = load_fully_qualified_func(fully_qual_func_name)
        # check if f is a function
        if (inspect.ismethod(f) or inspect.isfunction(f)):
            # get the class reference
            class_reference = get_class_that_defined_method(f)
            utils.debug("class for function: {}, {}".format(fully_qual_func_name, class_reference))

            # check if class is of TSV type
            if (class_reference == tsv.TSV):
                return True
            else:
                # check if bases classes 
                # check for base classes
                tsv_base_classes = list(filter(lambda t: t == tsv.TSV, class_reference.__bases__))
                if (len(tsv_base_classes) > 0):
                    return True
                else:
                    utils.error("verify_tsv_subclass_func: failed. fully_qual_func_name is not inheriting from TSV class: {}".format(fully_qual_func_name))
                    return False
        else:
            utils.error("verify_tsv_subclass_func: failed. fully_qual_func_name is not function: {}".format(fully_qual_func_name))
            return False
    except Exception as e:
        utils.error("verify_tsv_subclass_func: failed: {}".format(e))
        return False

# >>> type(lambda x: x+2)
# <class 'function'>
# >>> cluster_class_reflection.get_fully_qualified_name(lambda x: x+2)
# '__main__.<lambda>'
def is_lambda_func(func):
    # check type
    if (isinstance(func, (type(is_lambda_func))) and get_fully_qualified_name(func).endswith(".<lambda>")):
        return True
    else:
        return False

# check for builtin functions that dont have __dict__
# >>> cluster_class_reflection.get_fully_qualified_name(sum)
# 'builtins.sum'
def is_builtin_func(func):
    # check type
    if (type(func) in (type(sum), type(math.ceil))):
        return True
    else:
        return False

# TODO: return True
def has_valid_reflective_name(func):
    # if (get_fully_qualified_name(func).startswith("omigo_") == True):
    #     return True
    # else:
    #     return False
    return True


# get the hydra version of the class extension
def get_hydra_class(class_ref, fallback_modules = None):
    # get the module and class name
    module = inspect.getmodule(class_ref)
    module_name = module.__name__
    class_name = class_ref.__name__

    # construct a new class name
    hydra_class_name = "Hydra{}".format(class_name)

    # trace
    utils.trace("get_hydra_class: class_name: {}, module_name: {}, hydra_class_name: {}".format(class_name, module_name, hydra_class_name))

    # check if the existing module has the hydra class present
    for member in inspect.getmembers(module):
        # the member is a tuple (name, reference)
        utils.trace("get_hydra_class: member: {}".format(member[0]))

        # compare
        if (member[0] == hydra_class_name):
            # found the hydra class
            utils.info_once("get_hydra_class: found hydra version for class: {}: {}.{}".format(class_ref.__name__, module_name, member[0]))
            return member[1]

    # look into fallback_module if that is defined
    if (fallback_modules is not None):
        # iterate through all fallback modules
        for fallback_module in fallback_modules:
            # the hydra class doesnt exist in primary location. check cluster_tsv
            for member in inspect.getmembers(fallback_module):
                # the member is a tuple (name, reference)
                if (member[0] == hydra_class_name):
                    # found the hydra class
                    utils.info_once("get_hydra_class: found hydra version for class: {}: {}.{}".format(class_ref.__name__, fallback_module.__name__, member[0]))
                    return member[1]

    # raise exception that no hydra version found
    utils.warn("get_hydra_class: Not able to find the hydra version for class: {}".format(class_ref.__name__))
    return None

def load_class_ref_by_name(class_name):
    # split
    parts = class_name.split(".")

    # validation
    if (len(parts) < 2):
        raise Exception("Invalid class name: {}".format(class_name))

    # get module and class base name
    module_name = ".".join(parts[0:-1])
    class_base_name = parts[-1]

    # load module
    module = importlib.import_module(module_name)
    class_ref = getattr(module, class_base_name)

    # return
    return class_ref



from omigo_hydra import cluster_common_v2
from omigo_hydra import cluster_protocol_v2
from omigo_hydra import cluster_tsv
from omigo_hydra import cluster_arjun
from omigo_hydra import cluster_funcs
from omigo_hydra import cluster_shell_ext
from omigo_hydra.cluster_common_v2 import EntityType
from omigo_hydra.cluster_common_v2 import EntityState
from omigo_hydra.cluster_common_v2 import ClusterCapabilities
from omigo_hydra.cluster_common_v2 import ClusterPaths
from omigo_core import utils, timefuncs, etl
from omigo_ext import splunk_ext
# utils.enable_debug_mode()
import time
from omigo_core import tsv
import json
import datetime

# TODO: s3 and local are differing in deletion of all paths

ID_REF = None

# Entities
MASTER = None
RESOURCE_MANAGER = None
SWF_MANAGER = None
WF_MANAGER = None
JOB_MANAGER = None
TASK_MANAGER = None
WORKER = None
AGENT = None
DOUBLE_AGENT = None
INTELI_AGENT = None
CLIENT = None
SESSION = None
SWF = None
WF = None
JOB = None
TASK = None

# Protocols
MASTER_PROTOCOL = None
RESOURCE_MANAGER_PROTOCOL = None
SWF_MANAGER_PROTOCOL = None
WF_MANAGER_PROTOCOL = None
JOB_MANAGER_PROTOCOL = None
TASK_MANAGER_PROTOCOL = None
WORKER_PROTOCOL = None
AGENT_PROTOCOL = None
DOUBLE_AGENT_PROTOCOL = None
INTELI_AGENT_PROTOCOL = None
CLIENT_PROTOCOL = None
SESSION_PROTOCOL = None
SWF_PROTOCOL = None
WF_PROTOCOL = None
JOB_PROTOCOL = None
TASK_PROTOCOL = None

# Cluster Management Protocols
ALL_PROTOCOLS = None
CLUSTER_ADMIN = None
MASTER_ELECTION_PROTOCOL = None

def init(value):
    global ID_REF
    global MASTER
    global RESOURCE_MANAGER
    global SWF_MANAGER
    global WF_MANAGER
    global JOB_MANAGER
    global TASK_MANAGER
    global WORKER
    global AGENT
    global DOUBLE_AGENT
    global INTELI_AGENT
    global CLIENT
    global SESSION
    global SWF
    global WF
    global JOB
    global TASK

    global MASTER_PROTOCOL
    global RESOURCE_MANAGER_PROTOCOL
    global SWF_MANAGER_PROTOCOL
    global WF_MANAGER_PROTOCOL
    global JOB_MANAGER_PROTOCOL
    global TASK_MANAGER_PROTOCOL
    global WORKER_PROTOCOL
    global AGENT_PROTOCOL
    global DOUBLE_AGENT_PROTOCOL
    global INTELI_AGENT_PROTOCOL
    global CLIENT_PROTOCOL
    global SESSION_PROTOCOL
    global SWF_PROTOCOL 
    global WF_PROTOCOL
    global JOB_PROTOCOL
    global TASK_PROTOCOL

    global ALL_PROTOCOLS
    global CLUSTER_ADMIN
    global MASTER_ELECTION_PROTOCOL
 
    ts = timefuncs.get_utctimestamp_sec() 
    cluster_common_v2.ClusterIds.set_id(value)

    # Entities
    MASTER             = cluster_common_v2.ClusterEntityMaster.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.MASTER))
    RESOURCE_MANAGER   = cluster_common_v2.ClusterEntityResourceManager.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.RESOURCE_MANAGER))
    SWF_MANAGER        = cluster_common_v2.ClusterEntitySWFManager.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.SWF_MANAGER))
    WF_MANAGER         = cluster_common_v2.ClusterEntityWFManager.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.WF_MANAGER))
    JOB_MANAGER        = cluster_common_v2.ClusterEntityJobManager.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.JOB_MANAGER))
    TASK_MANAGER       = cluster_common_v2.ClusterEntityTaskManager.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.TASK_MANAGER))
    WORKER             = cluster_common_v2.ClusterEntityWorker.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.WORKER))
    AGENT              = cluster_common_v2.ClusterEntityAgent.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.AGENT))
    DOUBLE_AGENT       = cluster_common_v2.ClusterEntityDoubleAgent.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.DOUBLE_AGENT))
    INTELI_AGENT       = cluster_common_v2.ClusterEntityInteliAgent.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.INTELI_AGENT))
    CLIENT             = cluster_common_v2.ClusterEntityClient.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.CLIENT))
    SESSION            = cluster_common_v2.ClusterEntitySession.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.SESSION))

    # Protocols
    MASTER_PROTOCOL = cluster_protocol_v2.ClusterMasterProtocol(MASTER)
    RESOURCE_MANAGER_PROTOCOL = cluster_protocol_v2.ClusterResourceManagerProtocol(RESOURCE_MANAGER)
    SWF_MANAGER_PROTOCOL = cluster_protocol_v2.ClusterSWFManagerProtocol(SWF_MANAGER)
    WF_MANAGER_PROTOCOL  = cluster_protocol_v2.ClusterWFManagerProtocol(WF_MANAGER)
    JOB_MANAGER_PROTOCOL = cluster_protocol_v2.ClusterJobManagerProtocol(JOB_MANAGER)
    TASK_MANAGER_PROTOCOL = cluster_protocol_v2.ClusterTaskManagerProtocol(TASK_MANAGER)
    WORKER_PROTOCOL = cluster_protocol_v2.ClusterWorkerProtocol(WORKER)
    AGENT_PROTOCOL = cluster_protocol_v2.ClusterAgentProtocol(AGENT)
    DOUBLE_AGENT_PROTOCOL = cluster_protocol_v2.ClusterDoubleAgentProtocol(DOUBLE_AGENT)
    INTELI_AGENT_PROTOCOL = cluster_protocol_v2.ClusterInteliAgentProtocol(INTELI_AGENT)
    CLIENT_PROTOCOL = cluster_protocol_v2.ClusterClientProtocol(CLIENT)
    SESSION_PROTOCOL = cluster_protocol_v2.ClusterSessionProtocol(SESSION)
    # SWF_PROTOCOL = cluster_protocol_v2.ClusterSWFProtocol(SWF)
    # WF_PROTOCOL = cluster_protocol_v2.ClusterWFProtocol(WF)
    # JOB_PROTOCOL = cluster_protocol_v2.ClusterJobProtocol(JOB)
    # TASK_PROTOCOL = cluster_protocol_v2.ClusterTaskProtocol(TASK)

    ALL_PROTOCOLS = [MASTER_PROTOCOL, RESOURCE_MANAGER_PROTOCOL, SWF_MANAGER_PROTOCOL, WF_MANAGER_PROTOCOL, JOB_MANAGER_PROTOCOL, TASK_MANAGER_PROTOCOL, WORKER_PROTOCOL, AGENT_PROTOCOL, 
        DOUBLE_AGENT_PROTOCOL, INTELI_AGENT_PROTOCOL, CLIENT_PROTOCOL, SESSION_PROTOCOL]
        # SWF_PROTOCOL, WF_PROTOCOL, JOB_PROTOCOL, TASK_PROTOCOL]

    # admin and master election
    CLUSTER_ADMIN = cluster_protocol_v2.ClusterAdmin()
    MASTER_ELECTION_PROTOCOL = cluster_protocol_v2.ClusterMasterElectionProtocol(MASTER.entity_id)

# Run1 with initialization
def run1():
    for p in ALL_PROTOCOLS:
        p.initialize()

    # run election and refresh cache if needed
    if (MASTER_ELECTION_PROTOCOL.run_election()):
        MASTER_PROTOCOL.refresh_master_cache()

# Run1 with initialization
def run1_client():
    SESSION_PROTOCOL.initialize()
    CLIENT_PROTOCOL.initialize()

# Run2 with supervisor assignment and monitoring
def run2(n = 10, wait_sec = 5):
    for i in range(n):
        print("Running iteration: {}".format(i+1))
        for p in ALL_PROTOCOLS:
            if (p.entity.entity_type == cluster_common_v2.EntityType.MASTER):
                if (MASTER_ELECTION_PROTOCOL.run_election()):
                    MASTER_PROTOCOL.refresh_master_cache()
                # monitor supervisor
                p.monitor_incoming_for_supervisor()

            # run active child supervisors
            if (p.entity.entity_type in cluster_common_v2.EntityActiveSupervisorTypes):
                p.monitor_active_children()

            # run passive child supervisors
            if (p.entity.entity_type in cluster_common_v2.EntityPassiveSupervisorTypes):
                p.monitor_passive_children()

            # run tasks
            if (p.entity.entity_type in cluster_common_v2.EntityExecutorTypes):
                p.monitor_execution_tasks()

        utils.info("Sleeping for {} seconds".format(wait_sec))
        time.sleep(wait_sec)

# Debugging
def get_cluster_handler():
    return ClusterPaths.get_cluster_handler()

def get_local_cluster_handler():
    return ClusterPaths.get_local_cluster_handler()

def list_dirs(path):
    for f in get_cluster_handler().list_dirs(path):
        if (path != ""):
            print("{}/{}".format(path, f))
        else:
            print(f)

def get_dirs(path):
    dirs = []
    for f in get_cluster_handler().list_dirs(path):
        if (path != ""):
            dirs.append("{}/{}".format(path, f))
        else:
            dirs.append(f)

    return dirs

def print_paths(path):
    for f in sorted(get_cluster_handler().list_all_recursive(path)):
        if (path != ""):
            print("{}/{}".format(path, f))
        else:
            print(f)

def print_all_paths():
    for d in sorted(ClusterPaths.get_base_paths(), reverse = True):
        print("print_all_paths: {}".format(d))
        print_paths(d)

def remove_paths(path):
    get_cluster_handler().remove_dir_recursive(path, ignore_missing = True)

def read_json(path):
    return get_cluster_handler().read_json(path)

def read_most_recent_json(path):
    return get_cluster_handler().read_most_recent_json(path)

def read(path):
    return get_cluster_handler().read(path)

def remove_all_paths():
    for d in sorted(get_dirs(""), reverse = True):
        remove_paths(d)

    remove_paths("")

def read_tsv(path):
    return get_cluster_handler().read_tsv(path)

def write_tsv(path, xtsv):
    return get_cluster_handler().write_tsv(path, xtsv)

def read_workflow_input(wf_id, max_duration = 3*86400, sleep_sec = 3):
    # output file
    input_file = cluster_common_v2.ClusterPaths.get_entity_data_input_file(cluster_common_v2.EntityType.WF, wf_id, "input_id1", 0)

    # return
    return read_workflow_file_path(input_file, max_duration = max_duration, sleep_sec = sleep_sec)

def read_workflow_output(wf_id, max_duration = 3600, sleep_sec = 3):
    # output file
    output_file = cluster_common_v2.ClusterPaths.get_entity_data_output_file(cluster_common_v2.EntityType.WF, wf_id, "output_id1", 0)

    # return
    return read_workflow_file_path(output_file, max_duration = max_duration, sleep_sec = sleep_sec)

def read_workflow_file_path(file_path, max_duration = 3*86400, sleep_sec = 3):
    # check if the file exists, and run in a loop with sleep
    duration = 0
    found = False

    # run in a loop
    utils.info("Waiting for the file to be present: {} ...".format(file_path))
    while (duration < max_duration):
      if (get_cluster_handler().file_exists(file_path) == True):
          time.sleep(sleep_sec)
          found = True
          break
      else:
          utils.trace("File: {} doesnt exists yet. Sleeping for {} seconds".format(file_path, sleep_sec))
          time.sleep(sleep_sec)
          duration = duration + sleep_sec 

    # check if file was found
    if (found == True):
        try:
            xtsv = read_tsv(file_path)
        except Exception as e:
            utils.warn("Caught exception in reading file. Sleeping for {} seconds".format(sleep_sec))
            time.sleep(sleep_sec)
            xtsv = read_tsv(file_path)

        utils.info("Read file: {}, num_rows: {}".format(file_path, xtsv.num_rows()))
        return xtsv
    else:
        raise Exception("File not found: {}".format(file_path))

def scan_by_datetime_range(path, start_date_str, end_date_str, prefix):
    return etl.scan_by_datetime_range(get_cluster_handler().get_full_path(path), start_date_str, end_date_str, prefix)

#######################################################################################################################
# Script execution
#######################################################################################################################
def get_created_non_created_wf_ids(local_handler_flag = True, capabilities = []):
    # result
    created_wf_ids = []
    alive_wf_ids = []
    non_created_wf_ids = []
    capabilities = set(capabilities)
    handler = get_local_cluster_handler() if (local_handler_flag == True) else get_cluster_handler()

    # iterate over all states
    for state in cluster_common_v2.EntityState.get_all():
        # get all workflow ids
        ids = handler.list_dirs(ClusterPaths.get_entities_state_by_state(EntityType.WF, state))

        # read each workflow filter the workflows based on the capabilities matching requirements
        ids_allowed = []
        for id1 in ids:
            # read wf
            wf = cluster_common_v2.ClusterEntityWF.from_json(handler.read_most_recent_json(ClusterPaths.get_entity(EntityType.WF, id1)))
            requirements = set(wf.collect_requirements())

            # check if all requirements are met
            if (len(capabilities) == 0 or len(requirements.difference(capabilities)) == 0):
                ids_allowed.append(id1)

        # check for created
        if (state in (EntityState.CREATED)):
            for id1 in ids_allowed:
                created_wf_ids.append(id1)

        # check for alive
        if (state in (EntityState.ALIVE)):
            for id1 in ids_allowed:
                alive_wf_ids.append(id1)

        # check for finished.
        if (state in (EntityState.COMPLETED, EntityState.FAILED, EntityState.ABORTED)):
            for id1 in ids_allowed:
                non_created_wf_ids.append(id1)

    # debug
    utils.debug("get_created_non_created_wf_ids: created_wf_ids: {}, alive_wf_ids: {}, non_created_wf_ids: {}".format(created_wf_ids, alive_wf_ids, non_created_wf_ids))

    # effective ids
    effective_created_wf_ids = set(created_wf_ids).difference(set(non_created_wf_ids))
    effective_alive_wf_ids = set(alive_wf_ids).difference(set(non_created_wf_ids))
    effective_non_created_wf_ids = set(non_created_wf_ids)

    # return set difference
    return effective_created_wf_ids, effective_alive_wf_ids, effective_non_created_wf_ids 

def print_created_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = False, capabilities = capabilities)
    for id1 in created_wf_ids:
        print(id1)

def print_alive_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = False, capabilities = capabilities)
    for id1 in alive_wf_ids:
        print(id1)

def print_non_created_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = False, capabilities = capabilities)
    for id1 in non_created_wf_ids:
        print(id1)

def print_local_created_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = True, capabilities = capabilities)
    for id1 in created_wf_ids:
        print(id1)

def print_local_alive_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = True, capabilities = capabilities)
    for id1 in alive_wf_ids:
        print(id1)

def print_local_non_created_wf_ids(capabilities = []):
    created_wf_ids, alive_wf_ids, non_created_wf_ids = get_created_non_created_wf_ids(local_handler_flag = True, capabilities = capabilities)
    for id1 in non_created_wf_ids:
        print(id1)

# methods for external task execution
def print_shell_executor_scripts(wf_id):
    # read wf spec
    wf = cluster_common_v2.ClusterEntityWF.from_json(get_cluster_handler().read_most_recent_json(ClusterPaths.get_entity(EntityType.WF, wf_id)))

    # create wf protocol
    wf_protocol = cluster_protocol_v2.ClusterWFProtocol(wf)

    # read input. TODO: Use ctx api
    input_id = wf.entity_spec.input_ids[0]
    output_id = wf.entity_spec.output_ids[0]
    file_index = 0
    xinput = get_cluster_handler().read_tsv(ClusterPaths.get_entity_data_input_file(EntityType.WF, wf_id, input_id, file_index))

    # resolve the meta parameters for external task
    xinput_resolved = wf_protocol.resolve_external_task_meta_params(xinput, input_id, output_id, file_index)
    operations = wf_protocol.execute_single_round_get_operations(wf.entity_spec)

    # generate output
    xoutput = wf_protocol.execute_single_round(operations, xinput_resolved)

    # print the script
    print(utils.url_decode(xoutput.col_as_array("shell:output:url_encoded")[0]))

def print_wf_requirements(local_handler_flag = False):
    # get handler
    handler = get_local_cluster_handler() if (local_handler_flag == True) else get_cluster_handler()

    # get all wfs
    wf_ids = sorted(handler.list_dirs(ClusterPaths.get_entities_ids(EntityType.WF)))

    # iterate
    for wf_id in wf_ids:
        # use protocol class to get the current state
        xentity_state_protocol = cluster_protocol_v2.ClusterEntityStateProtocol(EntityType.WF, wf_id)
        effective_state = xentity_state_protocol.get_registered_state_resolved()

        # get wf details
        wf = cluster_common_v2.ClusterEntityWF.from_json(handler.read_most_recent_json(ClusterPaths.get_entity(EntityType.WF, wf_id)))
        requirements = wf.collect_requirements()

        # print
        utils.info("wf_id: {}, cur_state: {}, requirements: {}".format(wf_id, effective_state, requirements))


class EntityRunner:
    def __init__(self, ident):
        self.ident = ident
        self.protocol = None
        self.wait_sec = 5

    def setup(self):
        cluster_common_v2.ClusterIds.set_id(self.ident)

    def run_step(self):
        # run active child supervisors
        if (self.protocol.entity.entity_type in cluster_common_v2.EntityActiveSupervisorTypes):
            self.protocol.monitor_active_children()

        # run passive child supervisors
        if (self.protocol.entity.entity_type in cluster_common_v2.EntityPassiveSupervisorTypes):
            self.protocol.monitor_passive_children()

        # run tasks
        if (self.protocol.entity.entity_type in cluster_common_v2.EntityExecutorTypes):
            self.protocol.monitor_execution_tasks()

    def run(self):
        while (True):
            # run step
            self.run_step()

            # sleep
            utils.info("{}: sleeping for {} seconds".format(self.protocol.entity.entity_type, self.wait_sec))
            time.sleep(self.wait_sec)
            
class EntityMasterRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()

        # create protocols
        self.protocol = cluster_protocol_v2.ClusterMasterProtocol(cluster_common_v2.ClusterEntityMaster.new(cluster_common_v2.ClusterIds.get_entity_id(EntityType.MASTER)))
        self.election_protocol = cluster_protocol_v2.ClusterMasterElectionProtocol(self.protocol.entity.entity_id)

        # initialize
        self.protocol.initialize()

    def run(self):
        while (True):
            # run election
            if (self.election_protocol.run_election() == True):
                self.protocol.refresh_master_cache()

            # monitor incoming
            self.protocol.monitor_incoming_for_supervisor()

            # run base class
            self.run_step()

            # sleep
            utils.info("{}: Sleeping for {} seconds".format(self.protocol.entity.entity_type, self.wait_sec))
            time.sleep(self.wait_sec)

class EntityResourceManagerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterResourceManagerProtocol(cluster_common_v2.ClusterEntityResourceManager.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.RESOURCE_MANAGER)))
        self.protocol.initialize()

class EntityJobManagerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterJobManagerProtocol(cluster_common_v2.ClusterEntityJobManager.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.JOB_MANAGER)))
        self.protocol.initialize()

class EntityTaskManagerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterTaskManagerProtocol(cluster_common_v2.ClusterEntityTaskManager.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.TASK_MANAGER)))
        self.protocol.initialize()

class EntitySWFManagerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterSWFManagerProtocol(cluster_common_v2.ClusterEntitySWFManager.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.SWF_MANAGER)))
        self.protocol.initialize()

class EntityWFManagerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterWFManagerProtocol(cluster_common_v2.ClusterEntityWFManager.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.WF_MANAGER)))
        self.protocol.initialize()

class EntityAgentRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterAgentProtocol(cluster_common_v2.ClusterEntityAgent.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.AGENT)))
        self.protocol.initialize()

class EntityWorkerRunner(EntityRunner):
    def __init__(self, ident):
        super().__init__(ident)

    def setup(self):
        super().setup()
        self.protocol = cluster_protocol_v2.ClusterWorkerProtocol(cluster_common_v2.ClusterEntityWorker.new(
            cluster_common_v2.ClusterIds.get_entity_id(EntityType.WORKER)))
        self.protocol.initialize()

def run_master(ident = 1):
    entity_runner = EntityMasterRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_resource_manager(ident = 1):
    entity_runner = EntityResourceManagerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_job_manager(ident = 1):
    entity_runner = EntityJobManagerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_task_manager(ident = 1):
    entity_runner = EntityTaskManagerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_swf_manager(ident = 1):
    entity_runner = EntitySWFManagerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_wf_manager(ident = 1):
    entity_runner = EntityWFManagerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_agent(ident = 1):
    entity_runner = EntityAgentRunner(ident)
    entity_runner.setup()
    entity_runner.run()

def run_worker(ident = 1):
    entity_runner = EntityWorkerRunner(ident)
    entity_runner.setup()
    entity_runner.run()

import multiprocessing

if __name__ == "__main__":
    multiprocessing.freeze_support()
    entity_name = sys.argv[1]
    if (entity_name == "master"):
        run_master()




"""FileReader / FileWriter class"""

import gzip
import zipfile
import time

from omigo_hydra import s3_wrapper
from omigo_core import utils

# TODO: this is very inefficient. Use s3fs or write something better.
class FileWriter:
    """FileWriter class to write data into files"""
    def __init__(self, output_file_name, s3_region, aws_profile):
        self.output_file_name = output_file_name
        self.data = []
        self.s3_region = s3_region
        self.aws_profile = aws_profile

    def write(self, line):
        self.data.append(line)

    def close(self):
        output_zipf = None
        output_file = None
        if (self.output_file_name.startswith("s3://")):
            content = "".join(self.data)
            bucket_name, object_key = utils.split_s3_path(self.output_file_name)

            # persist
            s3_wrapper.put_file_with_text_content(bucket_name, object_key, content, self.s3_region, self.aws_profile)

            # TODO: wait for file to exist
            utils.warn_once("FileWriter: this has wait incorporated. this needs to be merged with hydra")
            num_attempts = 20
            wait_sec = 0.1
            for attempt in range(num_attempts):
                if (s3_wrapper.check_file_exists(self.output_file_name) == True):
                    return
                else:
                    utils.debug("FileWriter: file write doesnt finished yet. path: {}, waiting for {} seconds, attempt: {} / {}".format(self.output_file_name,
                        wait_sec, attempt, num_attempts))
                    time.sleep(wait_sec)

            # error
            if (s3_wrapper.check_file_exists(self.output_file_name) == False):
                raise Exception("FileWriter: write failed after {} attempts: {}".format(num_attempts, self.output_file_name))
        else:
            # construct output file
            if (self.output_file_name.endswith(".gz")):
                output_file = gzip.open(self.output_file_name, "wt")
                # write all the content
                for line in self.data:
                    output_file.write(line)
            elif (self.output_file_name.endswith(".zip")):
                output_zipf = zipfile.ZipFile(self.output_file_name, "w")
                output_file = output_zipf.open(self.output_file_name.split("/")[-1][0:-4], "w")
                # write all the content
                for line in self.data:
                    output_file.write(str.encode(line))
            else:
                output_file = open(self.output_file_name, "w")
                # write all the content
                for line in self.data:
                    output_file.write(line)

            # close
            if (output_file is not None):
                output_file.close()
            if (output_zipf is not None):
                output_zipf.close()

            # set data to None
            self.data = None

class TSVFileWriter:
    """FileWriter class to write data into files"""
    def __init__(self, s3_region, aws_profile):
        self.s3_region = s3_region
        self.aws_profile = aws_profile

    def save(self, xtsv, output_file_name):
        output_zipf = None
        output_file = None

        # initialize fs
        fs = s3io_wrapper.S3FSWrapper(s3_region = self.s3_region, aws_profile = self.aws_profile)

        # write
        content = xtsv.get_header() + "\n" + "\n".join(xtsv.get_data()) if (xtsv.num_rows() > 0) else xtsv.get_header()
        bucket_name, object_key = utils.split_s3_path(output_file_name)
        fs.write_text_file(bucket_name, object_key, content)

class FileReader:
    """FileReader class to read data files"""
    def __init__(self, input_file_name, s3_region, aws_profile):
        pass

    def read(self):
        pass

    def close(self):
        pass





from omigo_core import tsv, utils

class ShellExecutorBase:
    def __init__(self, template, default_props = {}):
        self.template = template
        self.default_props = default_props

    def generate_script(self, props = {}):
        # create base script
        script = self.template

        # create new props
        new_props = {}
        for k in self.default_props.keys():
            new_props[str(k)] = str(self.default_props[k])
        for k in props.keys():
            new_props[str(k)] = str(props[k])

        # create a new props with proper prefix
        for k in new_props.keys():
            k1 = "{" + "omigo.prop.{}".format(str(k)) + "}"
            value = str(new_props[k])
            script = script.replace(k1, value)

        # check if there are left over properties
        index = script.find("{omigo.prop.")
        if (index != -1):
            raise Exception("ShellExecutorBase: generate_script: failed to replace all properties: {}, {}".format(script[index:], new_props))
                   
        # return
        return script

class SparkJobShellExecutor(ShellExecutorBase):
    DEFAULT_TEMPLATE = """#!/bin/bash

# export certain conf parameters to be read through System.getProperty()
export OMIGO_ARJUN_INPUT_FILE="{omigo.arjun.input_file}"
export OMIGO_ARJUN_OUTPUT_FILE="{omigo.arjun.output_file}"
export HDFS_PATH="s3://proxy-some-path/"

spark-submit \\
  --class {omigo.prop.class} \\
  --executor-cores {omigo.prop.executor-cores} \\
  --master "{omigo.prop.master}" \\
  --deploy-mode "{omigo.prop.deploy-mode}" \\
  --executor-memory {omigo.prop.executor-memory} \\
  --num-executors {omigo.prop.num-executors} \\
  --driver-memory {omigo.prop.driver-memory} \\
  --driver-cores {omigo.prop.driver-cores} \\
  --conf "spark.yarn.maxAppAttempts={omigo.prop.spark.yarn.maxAppAttempts}" \\
  --conf "spark.driver.maxResultSize={omigo.prop.spark.driver.maxResultSize}" \\
  --conf "spark.executor.memoryOverhead={omigo.prop.spark.executor.memoryOverhead}" \\
  --conf "spark.speculation={omigo.prop.spark.speculation}" \\
  --conf "spark.dynamicAllocation.enabled={omigo.prop.spark.dynamicAllocation.enabled}" \\
  --conf "spark.sql.hive.caseSensitiveInferenceMode={omigo.prop.spark.sql.hive.caseSensitiveInferenceMode}" \\
  {omigo.prop.jar_location} \\
  {omigo.prop.command-line-args}

# write to local
hadoop fs -text "${HDFS_PATH}${OMIGO_ARJUN_OUTPUT_FILE}/*gz" | gzip -c > ${HYDRA_PATH}${OMIGO_ARJUN_OUTPUT_FILE}
"""

    DEFAULT_PROPS = {
        "executor-cores": "4",
        "master": "yarn",
        "deploy-mode": "client",
        "executor-memory": "4G",
        "num-executors": "10",
        "driver-memory": "16G",
        "driver-cores": "4",
        "spark.yarn.maxAppAttempts": "1",
        "spark.driver.maxResultSize": "4G",
        "spark.executor.memoryOverhead": "4G",
        "spark.speculation": "false",
        "spark.dynamicAllocation.enabled": "false",
        "spark.sql.hive.caseSensitiveInferenceMode": "NEVER_INFER"
    }

    def __init__(self, template = DEFAULT_TEMPLATE, default_props = DEFAULT_PROPS):
        super().__init__(template, default_props = default_props)

# Shell script executor for spark job. TODO: This is WIP
class SparkJobShellExecutorTSV(tsv.TSV):
    # init
    def __init__(self, header, data, url_encoded_template, props, jar_location):
        super().__init__(header, data)
        self.url_encoded_template = url_encoded_template
        self.props = props
        self.jar_location = jar_location

    # this generates the executable script as output
    def execute(self, main_class_col, command_line_template, output_col, prefix = "shell"):
        # method to be used for explode api
        def __execute_inner__(mp):
            # replace template with anything in the data
            template = utils.replace_template_props(mp, utils.url_decode(self.url_encoded_template))

            # set the new properties
            new_props = {}
            new_props["jar_location"] = self.jar_location
            new_props["class"] = mp[main_class_col]
            new_props["command-line-args"] = utils.replace_template_props(mp, command_line_template)

            # result
            result_mp = {}
            spark_shell_executor = SparkJobShellExecutor(template = template)
            result_mp[output_col + ":url_encoded"] = utils.url_encode(spark_shell_executor.generate_script(props = new_props))

            # return
            return [result_mp]

        # return
        return self \
            .explode(".*", __execute_inner__, prefix, collapse = False)
